@@ -1,7 +1,7 @@
 /*===========================================================================*
  * This file is part of the Abstract Library for Parallel Search (ALPS).     *
  *                                                                           *
- * ALPS is distributed under the Common Public License as part of the        *
+ * ALPS is distributed under the Eclipse Public License as part of the       *
  * COIN-OR repository (http://www.coin-or.org).                              *
  *                                                                           *
  * Authors:                                                                  *
@@ -17,7 +17,7 @@
  *          Matthew Saltzman, Clemson University                             *
  *                                                                           * 
  *                                                                           *
- * Copyright (C) 2001-2007, Lehigh University, Yan Xu, and Ted Ralphs.       *
+ * Copyright (C) 2001-2011, Lehigh University, Yan Xu, and Ted Ralphs.       *
  *===========================================================================*/
 
 #ifndef AlpsKnowledgeBrokerMPI_h_
@@ -72,8 +72,8 @@ class AlpsKnowledgeBrokerMPI : public AlpsKnowledgeBroker {
     /** The actual size of the cluster to which the process belongs. */
     int clusterSize_;
 
-    /** The reqested size of a cluster. */
-    int cluSize_;
+    /** The user reqested size of a cluster. */
+    int userClusterSize_;
 
     /** The local rank of the process in clusterComm_. */
     int clusterRank_;
@@ -95,6 +95,21 @@ class AlpsKnowledgeBrokerMPI : public AlpsKnowledgeBroker {
 
     /** Whether hub should also work as a worker. */
     bool hubWork_;
+
+    /** Send subtree request. */
+    MPI_Request subTreeRequest_;
+
+    /** Send model knoledge request. */
+    MPI_Request solRequestL_;
+    MPI_Request solRequestR_;
+
+    /** Send model knoledge request. */
+    MPI_Request modelKnowRequestL_;
+    MPI_Request modelKnowRequestR_;
+
+    /** Forward model knoledge request. */
+    MPI_Request forwardRequestL_;
+    MPI_Request forwardRequestR_;
     //@}
 
     /** @name Incumbent data
@@ -174,9 +189,6 @@ class AlpsKnowledgeBrokerMPI : public AlpsKnowledgeBroker {
 
     /** To record how many nodes processed for each hub */
     int* hubNodeProcesseds_;
-
-    /** To record how many nodes processed by the system. */
-    int systemNodeProcessed_;
     //@}
 
     /** @name Message counts
@@ -215,6 +227,15 @@ class AlpsKnowledgeBrokerMPI : public AlpsKnowledgeBroker {
      *
      */
     //@{
+    /** Master timer */
+    AlpsTimer masterTimer_;
+
+    /** Hub timer */
+    AlpsTimer hubTimer_;
+
+    /** Worker timer */
+    AlpsTimer workerTimer_;
+    
     /** The time spent in ramp up. */
     double rampUpTime_;
 
@@ -246,6 +267,9 @@ class AlpsKnowledgeBrokerMPI : public AlpsKnowledgeBroker {
     /** Indicate whether a worker need to as for work from its hub */
     bool blockAskForWork_;
 
+    /** Buffer attached to MPI when sharing generated knowledge. */
+    char *attachBuffer_;
+    
     /** Large message buffer. */
     char *largeBuffer_;
 
@@ -277,7 +301,7 @@ class AlpsKnowledgeBrokerMPI : public AlpsKnowledgeBroker {
 
     /** Temporily halt search */
     int haltSearch_;
-    
+
  protected:
 
     /** Initialize member data. */
@@ -314,7 +338,16 @@ class AlpsKnowledgeBrokerMPI : public AlpsKnowledgeBroker {
     void processMessages(char *&buffer,
                          MPI_Status &status, 
                          MPI_Request &request);
-    
+
+    /** Static load balancing: Root Initialization */
+    void rootInitMaster(AlpsTreeNode* root);
+    void rootInitHub();
+    void rootInitWorker();
+
+    /** Static load balancing: spiral */
+    void spiralMaster(AlpsTreeNode* root);
+    void spiralHub();
+    void spiralWorker();
     
     //------------------------------------------------------
 
@@ -334,6 +367,9 @@ class AlpsKnowledgeBrokerMPI : public AlpsKnowledgeBroker {
     /** Calculate the work quality and quantity on this process. */
     void updateWorkloadInfo();
 
+    virtual int getNumNodeLeftSystem()
+    { return static_cast<int>(systemWorkQuantity_); }
+    
     /** A worker donate its workload to the specified worker. */
     void donateWork(char*& buf, 
 		    int tag,
@@ -451,23 +487,14 @@ class AlpsKnowledgeBrokerMPI : public AlpsKnowledgeBroker {
     /** First receive the size and the contend of a node, then construct 
 	a subtree with this received node. */
     // NOTE: comm is hubComm_ or clusterComm_
-    void receiveSizeNode(int sender,
-			 MPI_Comm comm, 
-			 MPI_Status* status);
+    void receiveRampUpNode(int sender,
+			   MPI_Comm comm, 
+			   MPI_Status* status);
 
     /** Receive a subtree from the sender process and add it into
 	the subtree pool.*/
     void receiveSubTree(char*& buf, int sender, MPI_Status* status);
 
-    /** Send a buffer to the target process. */
-    // Only used in sendSubTree().
-    void sendBuf(char*& buf, 
-		 int size, 
-		 int position, 
-		 const int target, 
-		 const int tag, 
-		 MPI_Comm comm);
-
     /** Send the size and content of a buffer to the target process. */
     // NOTE: comm is hubComm_ or clusterComm_.
     void sendSizeBuf(char*& buf, 
@@ -480,8 +507,12 @@ class AlpsKnowledgeBrokerMPI : public AlpsKnowledgeBroker {
     /** Send the size and the content of the best node of a given subtree 
 	to the target process. */
     // NOTE: comm is hubComm_ or clusterComm_.
-    void sendSizeNode(const int target, MPI_Comm comm);
+    void sendRampUpNode(const int target, MPI_Comm comm);
 
+    /** Send a node from rampUpSubTree's node pool and generated model 
+        knowledge */
+    void sendNodeModelGen(int receiver, int doUnitWork);
+    
     /** Send a given subtree to the target process. */
     bool sendSubTree(const int target, AlpsSubTree*& st, int tag);
 
@@ -532,8 +563,14 @@ class AlpsKnowledgeBrokerMPI : public AlpsKnowledgeBroker {
     /** Receive error code and set solution status. */
     void recvErrorCode(char *& bufLarge);
     
+    /** Unpack the node, explore it and send load info to master. */
+    void spiralRecvProcessNode();
+    
+    /** Unpack msg and donate a node. */
+    void spiralDonateNode();
+    
  public:
-
+    
     /** Default construtor. 
         NOTE: must call initializeSearch() later. */
     AlpsKnowledgeBrokerMPI()
@@ -556,7 +593,7 @@ class AlpsKnowledgeBrokerMPI : public AlpsKnowledgeBroker {
     
     /** Destructor. */
     ~AlpsKnowledgeBrokerMPI();
-    
+
     /** Query the global rank of the process. */
     virtual int getProcRank() const { return globalRank_; }
 
@@ -594,10 +631,18 @@ class AlpsKnowledgeBrokerMPI : public AlpsKnowledgeBroker {
 
     /** @name Report search results. */
     //@{
-    /** The process queries the quality of the incumbent it stores. */
+    /** The process queries the quality of the incumbent this process stores. */
     virtual double getIncumbentValue() const {
+        double bestObj = ALPS_OBJ_MAX;
+        if (AlpsKnowledgeBroker::hasKnowledge(AlpsKnowledgeTypeSolution)) {
+	    bestObj = getBestKnowledge(AlpsKnowledgeTypeSolution).second;
+            if (incumbentValue_ > bestObj) {
+                return bestObj;
+            }
+        }
 	return incumbentValue_;
     }
+ 
     /** The master queries the quality of the best solution it knowns. */
     virtual double getBestQuality() const {
 	if (globalRank_ == masterRank_) {
@@ -613,6 +658,9 @@ class AlpsKnowledgeBrokerMPI : public AlpsKnowledgeBroker {
         }
     }
 
+    /** Get best estimalted quality in system. */
+    virtual double getBestEstimateQuality() { return systemWorkQuality_; }
+
     /** Master prints out the best solution that it knows. */
     virtual void printBestSolution(char* outputFile = 0) const;
 
